import ultrastarparser.versions as versions
import os
import shutil


class UltrastarReaderWriter:
    def __init__(self, txt_file_path) -> None:
        self.txt_file_path = txt_file_path
        self.songfolder = os.path.dirname(txt_file_path)
        self.encoding = "utf-8"

        self.song: versions.BaseUltrastarVersion

    def read(self) -> None:
        with open(
            self.txt_file_path, "r", encoding=self.encoding, errors="ignore"
        ) as f:
            lines = f.read()

        # Remove BOM if present. Fuck BOM
        lines = lines.removeprefix("\ufeff")

        self.song = versions.ultrastar_version_factory(lines)
        self.song.parse(lines)

    def write(self) -> None:
        song = ""

        for attribute in self.song.get_attributes():
            song += f"#{attribute}:{self.song._attributes[attribute]}\n"
        for line in self.song.get_body():
            song += f"{line}\n"

        with open(self.txt_file_path, "w", encoding=self.encoding) as f:
            f.write(song)

    def backup(self, backup_file_path: str, files: list[str] | None) -> None:
        """
        Backup the song to a backup folder.
        Removes existing files in the backup folder.

        :param backup_file_path: The path to the backup folder.
        :param files: Files to backup
        """
        if not os.path.exists(backup_file_path):
            os.makedirs(backup_file_path)
        elif len(os.listdir(backup_file_path)) != 0:
            os.rmdir(backup_file_path)
            os.makedirs(backup_file_path)

        if files is not None:
            shutil.copy(
                os.path.join(self.songfolder, self.txt_file_path), backup_file_path
            )
            return
        for file in files:
            shutil.copy(os.path.join(self.file), backup_file_path)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UltrastarReaderWriter):
            return False
        return self.song == other.song



from csv import DictWriter
import os
import json

from ultrastarparser import Song


class Library:
    def __init__(self, library_folder: str) -> None:
        self.library_folder = library_folder
        self.songs: list[Song] = []
        self.load_songs()

    def load_songs(self) -> None:
        self.songs.clear()
        for root, _, files in os.walk(self.library_folder):
            for file in files:
                if file.endswith(".txt"):
                    self.songs.append(Song(os.path.join(root, file)))

    def search(self, attribute: str, value: str) -> list[Song]:
        """
        Search for songs with the given attribute and value like 'ARTIST',
        'Bon Jovi' -> [UltraStarFile, ...]

        :param attribute: The attribute to search for.
        :param value: The value of the attribute to search for.
        :return: A list of UltraStarFile objects that match the search.
        """
        return [
            song
            for song in self.songs
            if (attr := song.get_attribute(attribute)) is not None
            and value.lower() in attr.lower()
        ]

    def least_common_divisor_attributes(self) -> list[str]:
        """
        Returns all attributes in use in the entire library.

        :return: A list of all attributes used in the library sorted to match
        the USDX format.
        """
        attributes = set()
        for ussong in self:
            attributes.update(ussong.get_attributes().keys())
        attributes = list(attributes)

        # TODO find a way to sort the attributes in a sensible way

        return attributes

    def export(
        self,
        path: str,
        export_format: str,
        attributes: list[str] | None = None,
    ) -> None:
        """
        Exports the library to a file. Exports only necessary attributes
        by default.

        :param path: The path to the json file.
        :param export_format: The format to export the library to. Can be
        ExportFormat.JSON or ExportFormat.CSV.
        :param attributes: A list of attributes to export like ['#ARTIST',
        '#TITLE']. If not provided, all necessary attributes are exported.
        :raises ValueError: If the export format is not supported.
        :raises FileNotFoundError: If the path to the file is invalid.
        :raises PermissionError: If the file cannot be written to.
        """
        if attributes is None:
            attributes = self.least_common_divisor_attributes()

        export_data = {}
        ussong: Song
        for ussong in self:
            song_data = {}
            for attribute in attributes:
                song_data[attribute] = ussong.get_attribute(attribute)
            export_data[ussong.common_name] = song_data

        match export_format.lower():
            case "json":
                with open(file=path, mode="w") as output_file:
                    json.dump(export_data, output_file, indent=4)

            case "csv":
                with open(file=path, mode="w") as output_file:
                    writer = DictWriter(
                        f=output_file, fieldnames=attributes, dialect="excel"
                    )
                    writer.writeheader()
                    for ussong in export_data.values():
                        writer.writerow(ussong)
            case _:
                raise ValueError(f"Export format '{export_format}' not supported.")

    def get_songs(self) -> list[Song]:
        return self.songs

    def get_song(self, index: int) -> Song:
        return self.songs[index]

    def __iter__(self) -> iter:
        return iter(self.songs)

    def __next__(self) -> Song:
        return next(self.songs)

    def __str__(self) -> str:
        return f"Library: {self.library_folder}"

    def __repr__(self) -> str:
        return f"Library({self.library_folder})"

    def __len__(self) -> int:
        return len(self.songs)

    def __getitem__(self, index: int) -> Song:
        return self.get_song(index)



from collections.abc import Callable
from typing import Optional
from functools import total_ordering


class AttributeMapping:
    def __init__(
        self,
        new_name: str,
        transform: Optional[Callable[[str, dict[str, str]], str]] | None = None,
    ) -> None:
        self.new_name = new_name
        self.transform = transform

    new_name: str
    # the callable takes the new value name and the attributes dict and returns the new value
    transform: Optional[Callable[[str, dict[str, str]], str]] | None = None


class VersionChangeError(ValueError):
    def __init__(
        self, version: "FormatVersion", message: str = "Unable to change version"
    ) -> None:
        self.version = version
        self.message = message
        super().__init__(f"{message}: {version}")


@total_ordering
class FormatVersion:
    major: int
    minor: int
    patch: int

    def __init__(self, version: str | tuple[int, int, int]) -> None:
        if isinstance(version, str):
            if version.startswith("v") or version.startswith("V"):
                version = version[1:]
            if "." not in version:
                raise ValueError("Invalid version format")
            self.major, self.minor, self.patch = map(int, version.split("."))
        else:
            self.major, self.minor, self.patch = version

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FormatVersion):
            return False
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.patch == other.patch
        )

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, FormatVersion):
            return False
        if self.major != other.major:
            return self.major < other.major
        if self.minor != other.minor:
            return self.minor < other.minor
        return self.patch < other.patch

    def __hash__(self) -> int:
        return hash((self.major, self.minor, self.patch))


@total_ordering
class BaseUltrastarVersion:
    def __init__(self) -> None:
        self._version: FormatVersion
        self._attributes: dict[str, str] = {}
        self._body: list[str] = []

        self.primary_audio_attributes: str | None

        self.required_attributes: list[str]
        self.optional_attributes: list[str]
        self.attribute_mappings: dict[str, dict[str, AttributeMapping]]

    def parse(self, file: str) -> None:
        attributes = {}
        body = []

        lines = file.splitlines()
        if len(lines) < 1:
            return
        headerFinished: bool = False
        for line in lines:
            if line.startswith("#") and not headerFinished:
                line = line.strip()
                key, value = line[1:].split(":", 1)
                key = key.upper()
                attributes[key.strip()] = value.strip()
            elif line == "" and not headerFinished:
                continue
            elif (line is None or line.isspace()) and headerFinished:
                body.append(line)
            elif line == "E":
                body.append(line)
                body.append("")
                break
            else:
                headerFinished = True
                body.append(line)

        self._set_attributes(attributes)
        self._set_body(body)

    def downgrade(self) -> "BaseUltrastarVersion":
        version_keys = list(versions.keys())
        current_index = version_keys.index(self._version)
        if current_index == 0:
            raise VersionChangeError(
                self.get_version(),
                "Cannot downgrade from version because it is the latest version.",
            )

        previous_version_key = version_keys[current_index - 1]
        previous_version_class: BaseUltrastarVersion = versions[previous_version_key]()

        attribute_mapping = self.attribute_mappings.get("downgrade", {})
        previous_version_class._attributes = {}
        for k, v in self._attributes.items():
            if k in attribute_mapping:
                mapping = attribute_mapping[k]
                new_key = mapping.new_name
                transform = mapping.transform
                previous_version_class._attributes[new_key] = (
                    transform(v) if transform else v
                )
            else:
                previous_version_class._attributes[k] = v

        previous_version_class._version = previous_version_key
        previous_version_class._attributes["VERSION"] = str(previous_version_key)
        previous_version_class._set_body(self._body)

        return previous_version_class

    def upgrade(self) -> "BaseUltrastarVersion":
        version_keys = list(versions.keys())
        current_index = version_keys.index(self._version)
        if current_index == len(version_keys) - 1:
            raise VersionChangeError(
                self.get_version(),
                "Cannot upgrade from version because it is the latest version.",
            )

        next_version_key = version_keys[current_index + 1]
        next_version_class: BaseUltrastarVersion = versions[next_version_key]()

        attribute_mapping = next_version_class.attribute_mappings.get("upgrade", {})
        next_version_class._attributes = {}
        next_version_class._version = next_version_key
        for k, v in self._attributes.items():
            if k in attribute_mapping:
                mapping = attribute_mapping[k]
                new_key = mapping.new_name
                transform = mapping.transform
                next_version_class._attributes[new_key] = (
                    transform(v) if transform else v
                )
            else:
                next_version_class._attributes[k] = v

        next_version_class._version = next_version_key
        next_version_class._attributes["VERSION"] = str(next_version_key)
        next_version_class._set_body(self._body)

        return next_version_class

    def _set_attributes(self, attributes: dict[str, str]) -> None:
        self._attributes = attributes

    def _set_body(self, body: list[str]) -> None:
        self._body = body

    def get_attributes(self) -> dict[str, str]:
        return self._attributes

    def get_attribute(self, attribute: str) -> str | None:
        if attribute in self._attributes:
            return self._attributes.get(attribute)
        return None

    def get_body(self) -> list[str]:
        return self._body

    def get_version(self) -> FormatVersion:
        return self._version

    def get_primary_audio(self) -> str | None:
        for key in self.primary_audio_attributes:
            if self.get_attribute(key) is not None:
                return self.get_attribute(key)
        else:
            return None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseUltrastarVersion):
            return False
        version_equal: bool = self.get_version() == other.get_version()
        attributes_equal: bool = self.get_attributes() == other.get_attributes()
        body_equal: bool = self.get_body() == other.get_body()

        return version_equal and attributes_equal and body_equal

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, BaseUltrastarVersion):
            return False
        return self.get_version() < other.get_version()


class UltrastarVersion010(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("0.1.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "MP3",
        "BPM",
    ]
    optional_attributes = []

    primary_audio_attributes = ["MP3"]


class UltrastarVersion020(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("0.2.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "MP3",
        "BPM",
    ]
    optional_attributes = [
        "GAP",
        "COVER",
        "BACKGROUND",
        "VIDEO",
        "VIDEOGAP",
        "GENRE",
        "EDITION",
        "CREATOR",
        "LANGUAGE",
        "YEAR",
        "START",
        "END",
        "PREVIEWSTART",
        "MEDLEYSTARTBEAT",
        "MEDLEYENDBEAT",
        "CALCMEDLEY",
        "DUETSINGERP1",
        "DUETSINGERP2",
        "P1",
        "P2",
        "COMMENT",
        "RESOLUTION",
        "NOTESGAP",
        "RELATIVE",
        "ENCODING",
    ]
    attribute_mappings = {}

    primary_audio_attributes = ["MP3"]


class UltrastarVersion030(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("0.3.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "MP3",
        "BPM",
    ]
    optional_attributes = [
        "GAP",
        "COVER",
        "BACKGROUND",
        "VIDEO",
        "VIDEOGAP",
        "GENRE",
        "EDITION",
        "CREATOR",
        "LANGUAGE",
        "YEAR",
        "START",
        "END",
        "PREVIEWSTART",
        "MEDLEYSTARTBEAT",
        "MEDLEYENDBEAT",
        "CALCMEDLEY",
        "DUETSINGERP1",
        "DUETSINGERP2",
        "P1",
        "P2",
        "COMMENT",
        "RESOLUTION",
        "NOTESGAP",
        "RELATIVE",
        "ENCODING",
    ]
    attribute_mappings = {}

    primary_audio_attributes = ["MP3"]


class UltrastarVersion100(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("1.0.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "MP3",
        "BPM",
    ]
    optional_attributes = [
        "GAP",
        "COVER",
        "BACKGROUND",
        "VIDEO",
        "VIDEOGAP",
        "GENRE",
        "EDITION",
        "CREATOR",
        "LANGUAGE",
        "YEAR",
        "START",
        "END",
        "PREVIEWSTART",
        "MEDLEYSTARTBEAT",
        "MEDLEYENDBEAT",
        "CALCMEDLEY",
        "P1",
        "P2",
        "COMMENT",
    ]
    attribute_mappings = {}

    primary_audio_attributes = ["MP3"]


class UltrastarVersion110(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("1.1.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "MP3",
        "AUDIO",
        "BPM",
    ]
    optional_attributes = [
        "VOCALS",
        "INSTRUMENTAL",
        "GAP",
        "COVER",
        "BACKGROUND",
        "VIDEO",
        "VIDEOGAP",
        "GENRE",
        "EDITION",
        "TAGS",
        "CREATOR",
        "LANGUAGE",
        "YEAR",
        "START",
        "END",
        "PREVIEWSTART",
        "MEDLEYSTARTBEAT",
        "MEDLEYENDBEAT",
        "CALCMEDLEY",
        "P1",
        "P2",
        "COMMENT",
        "PROVIDEDBY",
    ]
    attribute_mappings = {
        "upgrade": {
            "MP3": AttributeMapping(new_name="AUDIO", transform=None),
        },
        "downgrade": {
            "AUDIO": AttributeMapping(new_name="MP3", transform=None),
        },
    }

    primary_audio_attributes = ["MP3", "AUDIO"]


class UltrastarVersion120(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("1.1.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "MP3",
        "AUDIO",
        "BPM",
    ]
    optional_attributes = [
        "AUDIOURL",
        "VOCALS",
        "INSTRUMENTAL",
        "GAP",
        "COVER",
        "COVERURL",
        "BACKGROUND",
        "BACKGROUNDURL",
        "VIDEO",
        "VIDEOURL",
        "VIDEOGAP",
        "GENRE",
        "EDITION",
        "TAGS",
        "CREATOR",
        "LANGUAGE",
        "YEAR",
        "START",
        "END",
        "PREVIEWSTART",
        "MEDLEYSTARTBEAT",
        "MEDLEYENDBEAT",
        "CALCMEDLEY",
        "P1",
        "P2",
        "COMMENT",
        "PROVIDEDBY",
    ]
    attribute_mappings = {
        "upgrade": {
            "MP3": AttributeMapping(new_name="AUDIO", transform=None),
        },
        "downgrade": {
            "AUDIO": AttributeMapping(new_name="MP3", transform=None),
        },
    }

    primary_audio_attributes = ["MP3", "AUDIO"]


class UltrastarVersion200(BaseUltrastarVersion):
    _version: FormatVersion = FormatVersion("2.0.0")
    required_attributes = [
        "VERSION",
        "TITLE",
        "ARTIST",
        "AUDIO",
        "BPM",
    ]
    optional_attributes = [
        "GAP",
        "COVER",
        "BACKGROUND",
        "VIDEO",
        "VIDEOGAP",
        "GENRE",
        "EDITION",
        "CREATOR",
        "LANGUAGE",
        "YEAR",
        "START",
        "END",
        "PREVIEWSTART",
        "MEDLEYSTARTBEAT",
        "MEDLEYENDBEAT",
        "CALCMEDLEY",
        "P1",
        "P2",
        "COMMENT",
    ]
    attribute_mappings = {
        "upgrade": {
            "MP3": AttributeMapping(new_name="AUDIO", transform=None),
            "MEDLEYSTARTBEAT": AttributeMapping(
                new_name="MEDLEYSTART",
                transform=None,  # FIXME: transform
            ),
            "MEDLEYENDBEAT": AttributeMapping(
                new_name="MEDLEYEND",
                transform=None,  # FIXME: transform
            ),
        },
        "downgrade": {
            "AUDIO": AttributeMapping(new_name="MP3", transform=None),
            "MEDLEYSTART": AttributeMapping(new_name="MEDLEYSTARTBEAT", transform=None),
            "MEDLEYEND": AttributeMapping(new_name="MEDLEYENDBEAT", transform=None),
        },
    }

    primary_audio_attributes = ["MP3", "AUDIO"]


versions: dict[FormatVersion, BaseUltrastarVersion] = {
    FormatVersion("0.1.0"): UltrastarVersion010,
    FormatVersion("0.2.0"): UltrastarVersion020,
    FormatVersion("0.3.0"): UltrastarVersion030,
    FormatVersion("1.0.0"): UltrastarVersion100,
    FormatVersion("1.1.0"): UltrastarVersion110,
    FormatVersion("1.2.0"): UltrastarVersion120,
    FormatVersion("2.0.0"): UltrastarVersion200,
}
"""
Contains all supported Ultrastar file versions
"""


def _detect_version(attributes: dict[str, str]) -> FormatVersion:
    """
    Detects the Ultrastar file version based on the attributes present in the file.

    :param attributes: The attributes dict
    :return: The detected version
    """
    # Check if the version is explicitly defined. If so, and we support it, return it.
    if "VERSION" in attributes:
        version = FormatVersion(attributes["VERSION"])
        if version in versions.keys():
            return version

    best_version: FormatVersion = None
    max_optional_matches = -1

    for version_key in sorted(versions.keys(), reverse=True):
        version_class = versions.get(version_key)
        required_attrs = version_class.required_attributes
        optional_attrs = version_class.optional_attributes

        # Check if all required attributes are present
        if all(attr in attributes for attr in required_attrs):
            # Count matching optional attributes
            optional_matches = sum(1 for attr in optional_attrs if attr in attributes)

            # Select the version with the highest number of optional matches
            if optional_matches > max_optional_matches or (
                optional_matches == max_optional_matches
                and (best_version is None or version_key > best_version)
            ):
                best_version = version_key
                max_optional_matches = optional_matches

    if best_version is not None:
        return best_version

    # If no version was found, default to 1.0.0. This could be dangerous when upgrading/downgrading,
    # but we want to avoid crashing under any circumstances.
    return FormatVersion("1.0.0")


def ultrastar_version_factory(file: str) -> BaseUltrastarVersion:
    """
    Find a version for the Ultrastar file and return an instance of the corresponding class. If no version is found,
    return an instance of the 1.0.0 version.

    :param file: The Ultrastar file content. Not the file path.
    :return: An instance of the correct Ultrastar version class
    """
    attributes = {}
    lines = file.splitlines()
    for line in lines:
        line = line.strip()
        if line.startswith("#"):
            key, value = line[1:].split(":", 1)
            key = key.upper()
            attributes[key.strip()] = value.strip()
        elif line != "":
            break
    version = _detect_version(attributes)
    return versions.get(version)()



import ultrastarparser.io as io
import ultrastarparser.versions as versions
import os


class Song:
    """
    Represents an Ultrastar song.
    """

    def __init__(self, txt_file_path: str) -> None:
        """
        :param txt_file_path: Path to the Ultrastar song file. The folder
        containing the song file is considered the song folder and should not
        contain any more ultrastar text files.
        """
        self._reader_writer = io.UltrastarReaderWriter(txt_file_path)
        self.parse()

        self.songfolder = os.path.dirname(txt_file_path)

        artist = self.get_attribute("ARTIST")
        title = self.get_attribute("TITLE")
        self.common_name = (
            f"{artist} - {title}" if artist is not None and title is not None else None
        )

    def parse(self) -> None:
        """
        Parse the song file. This is done automatically when the song is created.
        Reparsing the song before flushing changes resets the changes made to the
        song.
        """
        self._reader_writer.read()

    def get_attribute(self, attribute: str) -> str | None:
        """
        Get an attribute from the song.

        :param attribute: The attribute to get.
        """
        return self._reader_writer.song.get_attribute(attribute.upper())

    def get_attributes(self) -> dict[str, str]:
        """
        Get all attributes from the song.

        :return: A dictionary of all attributes in the song.
        """
        return self._reader_writer.song.get_attributes()

    def set_attribute(self, attribute: str, value: str) -> None:
        """
        Set an attribute in the song.

        :param attribute: The attribute to set.
        :param value: The value to set the attribute to.
        """
        self._reader_writer.song._attributes[attribute.upper()] = value

    def get_songtext(self) -> list[str]:
        """
        Get the song text of the song.

        :return: The song text.
        """
        return self._reader_writer.song.get_body()

    def get_version(self) -> str:
        """
        Get the version of the ultrastar song file.

        :return: The version of the ultrastar song file.
        """
        return str(self._reader_writer.song.get_version())

    def set_version(self, dest_version: str) -> None:
        """
        Set the version of the ultrastar song file. Upgrades or downgrades the
        song to the desired version. This may result in data loss if the version
        to be changed to or an intermediate version removes attributes or changes
        the format of the song.

        This operation is not reversible. Upgrading a song to a higher version and
        then downgrading it to the original version might not result in the same
        song as the original.

        Some attributes' values will be lost. For example, version 1.0.0 depreciates
        the "DUETSINGERPX" attributes. However, we cannot naively copy these values
        to the new "PX" attributes, as those also existed before version 1.0.0. In this
        case, the values are lost.

        :param dest_version: The version to change the song to.
        """
        dest_version = versions.FormatVersion(dest_version)
        current_version = versions.FormatVersion(self.get_version())
        if dest_version == current_version:
            return
        available_versions = versions.versions
        if dest_version not in available_versions.keys():
            raise ValueError(
                f"This version doesn't exist or is unsupported: {dest_version}"
            )
        upgrade = dest_version > current_version

        # Upgrade or downgrade version until it matches the desired version
        while current_version != dest_version:
            if upgrade:
                try:
                    self._reader_writer.song = self._reader_writer.song.upgrade()
                except versions.VersionChangeError:
                    return
            else:
                try:
                    self._reader_writer.song = self._reader_writer.song.downgrade()
                except versions.VersionChangeError:
                    return

    def get_primary_audio(self) -> str | None:
        """
        Get the primary audio file of the song.

        :return: The primary audio file of the song.
        """
        return self._reader_writer.song.get_primary_audio()

    def flush(self) -> None:
        """
        Flush changes to the song file to the file system. Until this method is
        called, changes are only stored in memory.
        """
        self._reader_writer.write()

    def backup(self, backup_folder: str) -> None:
        """
        Backup the song file to a folder.

        :param backup_folder: The folder to backup the song to.
        """
        raise NotImplementedError("This method is not implemented yet.")

    def __str__(self) -> str:
        return self.common_name

    def __repr__(self) -> str:
        return f"Song: {self.common_name}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Song):
            return False
        return self._reader_writer == other._reader_writer

    def __hash__(self) -> int:
        return hash(self._reader_writer)
